@@ -4,8 +4,9 @@ import sys
 import os
 from ctypes import *
 import _ctypes
+from typing import Union, IO, List
 
-__version__ = '0.2.27'
+__version__ = '0.3.2'
 
 # experimental
 plot_library = 'matplotlib'  # 'plotly'
@@ -61,7 +62,7 @@ else:
 platform_tuple = architecture + '-' + system
 
 
-def supported_platforms(filename):
+def supported_platforms(filename: Union[str, IO]):
     """ Get the platforms supported by the FMU without extracting it
 
     Parameters:
@@ -71,10 +72,8 @@ def supported_platforms(filename):
         platforms   a list of supported platforms supported by the FMU
     """
 
-    from .util import _is_string
-
     # get the files within the FMU
-    if _is_string(filename) and os.path.isdir(filename):  # extracted FMU
+    if isinstance(filename, str) and os.path.isdir(filename):  # extracted FMU
         names = []
         for dirpath, _, filenames in os.walk(filename):
             for name in filenames:
@@ -208,15 +207,16 @@ def extract(filename, unzipdir=None, include=None):
     return unzipdir
 
 
-def dump(filename):
+def dump(filename: Union[str, IO], causalities: List[str] = ['input', 'output']):
     """ Print the model information and variables of an FMU
 
     Parameters:
-        filename    filename of the FMU
+        filename     filename of the FMU
+        causalities  the causalities of the variables to include
     """
 
     from .util import fmu_info
-    print(fmu_info(filename))
+    print(fmu_info(filename=filename, causalities=causalities))
 
 
 # make the functions available in the fmpy module
