@@ -1,15 +1,19 @@
 # noinspection PyPep8
 
 import shutil
+
+from fmpy.model_description import ModelDescription
+
 from .fmi1 import *
-from .fmi1 import _FMU1
+from .fmi1 import _FMU1, _FMU
 from .fmi2 import *
 from .fmi2 import _FMU2
 from . import fmi3
 from . import extract
-from .util import auto_interval, _is_string
+from .util import auto_interval
 import numpy as np
 from time import time as current_time
+from typing import Union, Any, Dict, Sequence, Callable
 
 # absolute tolerance for equality when comparing two floats
 eps = 1e-13
@@ -176,12 +180,13 @@ class Recorder(object):
 class Input(object):
     """ Helper class that sets the input to the FMU """
 
-    def __init__(self, fmu, modelDescription, signals):
+    def __init__(self, fmu, modelDescription, signals, set_input_derivatives=False):
         """
         Parameters:
-            fmu               the FMU instance
-            modelDescription  the model description instance
-            signals           a structured numpy array that contains the input
+            fmu                    the FMU instance
+            modelDescription       the model description instance
+            signals                a structured numpy array that contains the input
+            set_input_derivatives  calculate and set the input derivatives
 
         Example:
 
@@ -204,11 +209,11 @@ class Input(object):
         # find events
         self.t_events = Input.findEvents(signals, modelDescription)
 
+        self.set_input_derivatives = set_input_derivatives
+
         is_fmi1 = isinstance(fmu, _FMU1)
         is_fmi2 = isinstance(fmu, _FMU2)
 
-        self.set_input_derivatives = is_fmi2 and modelDescription.coSimulation and modelDescription.coSimulation.canInterpolateInputs
-
         setters = dict()
 
         # get the setters
@@ -366,6 +371,8 @@ class Input(object):
         # check for event
         if time == t[i0] and i0 < len(t) - 1 and t[i0] == t[i0 + 1]:
 
+            der_v = np.zeros((table.shape[0],))
+
             if after_event:
                 # take the value after the event
                 while i0 < len(t) - 1 and t[i0] == t[i0 + 1]:
@@ -375,15 +382,15 @@ class Input(object):
                     v1 = table[:, i0 + 1]
                     t0 = t[i0]
                     t1 = t[i0 + 1]
-                    der_v = (v1 - v0) / (t1 - t0)
-                else:
-                    der_v = np.zeros((table.shape[0],))
+                    if not discrete:
+                        der_v = (v1 - v0) / (t1 - t0)
             else:
                 v0 = table[:, i0 - 1]
                 v1 = table[:, i0]
                 t0 = t[i0 - 1]
                 t1 = t[i0]
-                der_v = (v1 - v0) / (t1 - t0)
+                if not discrete:
+                    der_v = (v1 - v0) / (t1 - t0)
 
             values = table[:, i0]
             return values, der_v
@@ -481,7 +488,7 @@ def apply_start_values(fmu, model_description, start_values, apply_default_start
 
         # convert the type
         if variable.shape:
-            if _is_string(value):
+            if isinstance(value, str):
                 value = value.split()
             value = list(map(lambda e: variable._python_type(e), value))
             if len(value) != np.prod(variable.shape):
@@ -552,72 +559,90 @@ class ForwardEuler(object):
 
 
 def simulate_fmu(filename,
-                 validate=True,
-                 start_time=None,
-                 stop_time=None,
-                 solver='CVode',
-                 step_size=None,
-                 relative_tolerance=None,
-                 output_interval=None,
-                 record_events=True,
-                 fmi_type=None,
-                 start_values={},
-                 apply_default_start_values=False,
-                 input=None,
-                 output=None,
-                 timeout=None,
-                 debug_logging=False,
-                 visible=False,
-                 logger=None,
-                 fmi_call_logger=None,
-                 step_finished=None,
-                 model_description=None,
-                 fmu_instance=None):
+                 validate: bool = True,
+                 start_time: Union[float, str] = None,
+                 stop_time: Union[float, str] = None,
+                 solver: str ='CVode',
+                 step_size: Union[float, str] = None,
+                 relative_tolerance: Union[float, str] = None,
+                 output_interval: Union[float, str] = None,
+                 record_events: bool = True,
+                 fmi_type: str = None,
+                 start_values: Dict[str, Any] = {},
+                 apply_default_start_values: bool = False,
+                 input: np.ndarray = None,
+                 output: Sequence[str] = None,
+                 timeout: Union[float, str] = None,
+                 debug_logging: bool = False,
+                 visible: bool = False,
+                 logger: Callable = None,
+                 fmi_call_logger: Callable[[str], None] = None,
+                 step_finished: Callable[[float, Recorder], bool] = None,
+                 model_description: ModelDescription = None,
+                 fmu_instance: _FMU = None,
+                 set_input_derivatives: bool = False,
+                 remote_platform: str = 'auto') -> SimulationResult:
     """ Simulate an FMU
 
     Parameters:
-        filename            filename of the FMU or directory with extracted FMU
-        validate            validate the FMU
-        start_time          simulation start time (None: use default experiment or 0 if not defined)
-        stop_time           simulation stop time (None: use default experiment or start_time + 1 if not defined)
-        solver              solver to use for model exchange ('Euler' or 'CVode')
-        step_size           step size for the 'Euler' solver
-        relative_tolerance  relative tolerance for the 'CVode' solver and FMI 2.0 co-simulation FMUs
-        output_interval     interval for sampling the output
-        record_events       record outputs at events (model exchange only)
-        fmi_type            FMI type for the simulation (None: determine from FMU)
-        start_values        dictionary of variable name -> value pairs
+        filename               filename of the FMU or directory with extracted FMU
+        validate               validate the FMU
+        start_time             simulation start time (None: use default experiment or 0 if not defined)
+        stop_time              simulation stop time (None: use default experiment or start_time + 1 if not defined)
+        solver                 solver to use for model exchange ('Euler' or 'CVode')
+        step_size              step size for the 'Euler' solver
+        relative_tolerance     relative tolerance for the 'CVode' solver and FMI 2.0 co-simulation FMUs
+        output_interval        interval for sampling the output
+        record_events          record outputs at events (model exchange only)
+        fmi_type               FMI type for the simulation (None: determine from FMU)
+        start_values           dictionary of variable name -> value pairs
         apply_default_start_values  apply the start values from the model description
-        input               a structured numpy array that contains the input (see :class:`Input`)
-        output              list of variables to record (None: record outputs)
-        timeout             timeout for the simulation
-        debug_logging       enable the FMU's debug logging
-        visible             interactive mode (True) or batch mode (False)
-        fmi_call_logger     callback function to log FMI calls
-        logger              callback function passed to the FMU (experimental)
-        step_finished       callback to interact with the simulation (experimental)
-        model_description   the previously loaded model description (experimental)
-        fmu_instance        the previously instantiated FMU (experimental)
-
+        input                  a structured numpy array that contains the input (see :class:`Input`)
+        output                 list of variables to record (None: record outputs)
+        timeout                timeout for the simulation
+        debug_logging          enable the FMU's debug logging
+        visible                interactive mode (True) or batch mode (False)
+        fmi_call_logger        callback function to log FMI calls
+        logger                 callback function passed to the FMU (experimental)
+        step_finished          callback to interact with the simulation (experimental)
+        model_description      the previously loaded model description (experimental)
+        fmu_instance           the previously instantiated FMU (experimental)
+        set_input_derivatives  set the input derivatives (FMI 2.0 Co-Simulation only)
+        remote_platform        platform to use for remoting server ('auto': determine automatically if current platform
+                               is not supported, None: no remoting; experimental)
     Returns:
         result              a structured numpy array that contains the result
     """
 
-    from fmpy import supported_platforms, platform_tuple
+    from fmpy import supported_platforms
     from fmpy.model_description import read_model_description
+    from fmpy.util import has_wine64, has_wsl, can_simulate
 
     platforms = supported_platforms(filename)
 
-    # use 32-bit DLL remoting
-    use_remoting = platform == 'win64' and 'win64' not in platforms and 'win32' in platforms
+    if fmu_instance is None and platform not in platforms and remote_platform is None:
+        raise Exception(f"The current platform ({platform}) is not supported by the FMU.")
 
-    if fmu_instance is None and platform not in platforms and not use_remoting:
-        raise Exception("The current platform (%s) is not supported by the FMU." % platform)
+    can_sim, remote_platform = can_simulate(platforms, remote_platform)
+
+    if not can_sim:
+        raise Exception(f"The FMU cannot be simulated on the current platform ({platform}).")
+
+    # if platform not in platforms and remote_platform == 'auto':
+    #     if platform == 'win64' and 'win32' in platforms:
+    #         remote_platform = 'win32'
+    #     elif platform == 'win64' and 'linux64' in platforms and has_wsl():
+    #         remote_platform = 'linux64'
+    #     elif platform == 'linux64' and 'win64' in platforms and has_wine64():
+    #         remote_platform = 'win64'
+    #     else:
+    #         raise Exception(f"The FMU cannot be simulated on the current platform ({platform}).")
+    #
+    # if platform not in platforms and remote_platform not in platforms:
+    #     raise Exception(f"The current platform ({platform}) is not supported by the FMU.")
 
     if model_description is None:
         model_description = read_model_description(filename, validate=validate)
-    else:
-        model_description = model_description
 
     if fmi_type is None:
         if fmu_instance is not None:
@@ -664,28 +689,96 @@ def simulate_fmu(filename,
         unzipdir = filename
         tempdir = None
     else:
-        required_paths = ['resources', 'binaries/' + platform, 'binaries/' + platform_tuple]
-        if use_remoting:
-            required_paths.append('binaries/win32')
-        tempdir = extract(filename, include=lambda n: n.startswith(tuple(required_paths)))
+        required_paths = ['resources', 'binaries/']
+        if remote_platform:
+            required_paths.append(os.path.join('binaries', remote_platform))
+        tempdir = extract(filename, include=None if remote_platform else lambda n: n.startswith(tuple(required_paths)))
         unzipdir = tempdir
 
-    if use_remoting:
-        # start 32-bit server
-        from subprocess import Popen
-        server_path = os.path.dirname(__file__)
-        server_path = os.path.join(server_path, 'remoting', 'server.exe')
+    if remote_platform is None:
+
+        server = None
+        library_path = None
+
+    else:
+
+        import subprocess
+
         if fmi_type == 'ModelExchange':
             model_identifier = model_description.modelExchange.modelIdentifier
         else:
             model_identifier = model_description.coSimulation.modelIdentifier
-        dll_path = os.path.join(unzipdir, 'binaries', 'win32', model_identifier + '.dll')
-        server = Popen([server_path, dll_path])
-    else:
-        server = None
+
+        fmpy_dir = os.path.dirname(__file__)
+
+        if platform == 'win64':
+
+            if remote_platform == 'win32':
+
+                server_path = os.path.join(fmpy_dir, 'remoting', 'win32', 'server.exe')
+                dll_path = os.path.join(unzipdir, 'binaries', 'win32', model_identifier + '.dll')
+
+                library_path = os.path.join(fmpy_dir, 'remoting', 'win64', 'client.dll')
+                server = subprocess.Popen([server_path, dll_path])
+
+            elif remote_platform == 'linux64':
+
+                server_path = os.path.join(fmpy_dir, 'remoting', 'linux64', 'server').replace('\\', '/')
+                process = subprocess.run(['wsl', 'wslpath', server_path], capture_output=True, check=True)
+                server_path = process.stdout.decode("utf-8") .strip()
+
+                so_path = os.path.join(unzipdir, 'binaries', 'linux64', model_identifier + '.so').replace('\\', '/')
+                process = subprocess.run(['wsl', 'wslpath', so_path], capture_output=True, check=True)
+                so_path = process.stdout.decode("utf-8") .strip()
+
+                library_path = os.path.join(fmpy_dir, 'remoting', 'win64', 'client.dll')
+                server = subprocess.Popen(['wsl', server_path, so_path])
+
+            elif remote_platform == 'win64':
+
+                server_path = os.path.join(fmpy_dir, 'remoting', 'win64', 'server.exe')
+                dll_path = os.path.join(unzipdir, 'binaries', 'win64', model_identifier + '.dll')
+
+                library_path = os.path.join(fmpy_dir, 'remoting', 'win64', 'client.dll')
+                server = subprocess.Popen([server_path, dll_path])
+
+            else:
+
+                raise Exception(f"The remote platform {remote_platform} is not supported on the current platform ({platform}).")
+
+        elif platform == 'linux64':
+
+            if remote_platform == 'win32':
+
+                server_path = os.path.join(fmpy_dir, 'remoting', 'win32', 'server.exe')
+                dll_path = os.path.join(unzipdir, 'binaries', 'win32', model_identifier + '.dll')
+                library_path = os.path.join(fmpy_dir, 'remoting', 'linux64', 'client.so')
+                server = subprocess.Popen(['wine', server_path, dll_path])
+
+            elif remote_platform == 'win64':
+
+                server_path = os.path.join(fmpy_dir, 'remoting', 'win64', 'server.exe')
+                dll_path = os.path.join(unzipdir, 'binaries', 'win64', model_identifier + '.dll')
+                library_path = os.path.join(fmpy_dir, 'remoting', 'linux64', 'client.so')
+                server = subprocess.Popen(['wine64', server_path, dll_path])
+
+            elif remote_platform == 'linux64':
+
+                server_path = os.path.join(fmpy_dir, 'remoting', 'linux64', 'server')
+                dll_path = os.path.join(unzipdir, 'binaries', 'linux64', model_identifier + '.so')
+                library_path = os.path.join(fmpy_dir, 'remoting', 'linux64', 'client.so')
+                server = subprocess.Popen([server_path, dll_path])
+
+            else:
+
+                raise Exception(f"The remote platform {remote_platform} is not supported on the current platform ({platform}).")
+
+        else:
+
+            raise Exception(f"Remoting is not supported on the current platform ({platform}).")
 
     if fmu_instance is None:
-        fmu = instantiate_fmu(unzipdir, model_description, fmi_type, visible, debug_logging, logger, fmi_call_logger, use_remoting)
+        fmu = instantiate_fmu(unzipdir, model_description, fmi_type, visible, debug_logging, logger, fmi_call_logger, library_path)
     else:
         fmu = fmu_instance
 
@@ -693,13 +786,14 @@ def simulate_fmu(filename,
     if fmi_type == 'ModelExchange':
         result = simulateME(model_description, fmu, start_time, stop_time, solver, step_size, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, record_events, timeout, step_finished)
     elif fmi_type == 'CoSimulation':
-        result = simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, timeout, step_finished)
+        result = simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, timeout, step_finished, set_input_derivatives)
 
     if fmu_instance is None:
         fmu.freeInstance()
 
     if server is not None:
-        server.kill()
+        server.terminate()
+        server.wait()
 
     # clean up
     if tempdir is not None:
@@ -708,7 +802,7 @@ def simulate_fmu(filename,
     return result
 
 
-def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, debug_logging=False, logger=None, fmi_call_logger=None, use_remoting=False):
+def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, debug_logging=False, logger=None, fmi_call_logger=None, library_path=None):
     """
     Create an instance of fmpy.fmi1._FMU (see simulate_fmu() for documentation of the parameters).
     """
@@ -721,8 +815,8 @@ def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, d
         'fmiCallLogger': fmi_call_logger
     }
 
-    if use_remoting:
-        fmu_args['libraryPath'] = os.path.join(os.path.dirname(__file__), 'remoting', 'client.dll')
+    if library_path:
+        fmu_args['libraryPath'] = library_path
 
     if logger is None:
         logger = printLogMessage
@@ -823,7 +917,7 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
     if is_fmi1:
         fmu.setTime(time)
     elif is_fmi2:
-        fmu.setupExperiment(startTime=start_time)
+        fmu.setupExperiment(startTime=start_time, stopTime=stop_time)
 
     input = Input(fmu, model_description, input_signals)
 
@@ -869,21 +963,21 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
 
     elif is_fmi3:
 
-        fmu.enterInitializationMode(startTime=start_time)
+        fmu.enterInitializationMode(startTime=start_time, stopTime=stop_time)
         input.apply(time)
         fmu.exitInitializationMode()
 
-        newDiscreteStatesNeeded = True
+        discreteStatesNeedUpdate = True
         terminateSimulation = False
 
-        while newDiscreteStatesNeeded and not terminateSimulation:
+        while discreteStatesNeedUpdate and not terminateSimulation:
             # update discrete states
-            (newDiscreteStatesNeeded,
+            (discreteStatesNeedUpdate,
              terminateSimulation,
              nominalsOfContinuousStatesChanged,
              valuesOfContinuousStatesChanged,
              nextEventTimeDefined,
-             nextEventTime) = fmu.newDiscreteStates()
+             nextEventTime) = fmu.updateDiscreteStates()
 
         fmu.enterContinuousTimeMode()
 
@@ -893,7 +987,7 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
         'nz': model_description.numberOfEventIndicators,
         'get_x': fmu.getContinuousStates,
         'set_x': fmu.setContinuousStates,
-        'get_dx': fmu.getDerivatives,
+        'get_dx': fmu.getContinuousStateDerivatives if is_fmi3 else fmu.getDerivatives,
         'get_z': fmu.getEventIndicators,
         'input': input
     }
@@ -928,6 +1022,8 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
 
     t_next = start_time
 
+    n_fixed_steps = 0
+
     # simulation loop
     while time < stop_time:
 
@@ -935,10 +1031,10 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
             break
 
         if fixed_step:
-            if time + step_size < stop_time + eps:
-                t_next = time + step_size
-            else:
+            t_next = start_time + n_fixed_steps * step_size
+            if t_next > stop_time:
                 break
+            n_fixed_steps += 1
         else:
             if time + eps >= t_next:  # t_next has been reached
                 # integrate to the next grid point
@@ -965,6 +1061,8 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
             state_event, roots_found, time = solver.step(time, t_next)
         else:
             # skip
+            state_event = False
+            roots_found = []
             time = t_next
 
         # set the time
@@ -1045,7 +1143,7 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
                      nominalsOfContinuousStatesChanged,
                      valuesOfContinuousStatesChanged,
                      nextEventTimeDefined,
-                     nextEventTime) = fmu.newDiscreteStates()
+                     nextEventTime) = fmu.updateDiscreteStates()
 
                 if terminateSimulation:
                     break
@@ -1072,7 +1170,10 @@ def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_
     return recorder.result()
 
 
-def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, timeout, step_finished):
+def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, timeout, step_finished, set_input_derivatives):
+
+    if set_input_derivatives and not model_description.coSimulation.canInterpolateInputs:
+        raise Exception("Parameter set_input_derivatives is True but the FMU cannot interpolate inputs.")
 
     if output_interval is None:
         output_interval = auto_interval(stop_time - start_time)
@@ -1083,9 +1184,9 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
     is_fmi2 = model_description.fmiVersion == '2.0'
 
     if is_fmi2:
-        fmu.setupExperiment(tolerance=relative_tolerance, startTime=start_time)
+        fmu.setupExperiment(tolerance=relative_tolerance, startTime=start_time, stopTime=stop_time)
 
-    input = Input(fmu=fmu, modelDescription=model_description, signals=input_signals)
+    input = Input(fmu=fmu, modelDescription=model_description, signals=input_signals, set_input_derivatives=set_input_derivatives)
 
     time = start_time
 
@@ -1100,16 +1201,16 @@ def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance
         input.apply(time)
         fmu.exitInitializationMode()
     else:
-        fmu.enterInitializationMode(tolerance=relative_tolerance, startTime=start_time)
+        fmu.enterInitializationMode(tolerance=relative_tolerance, startTime=start_time, stopTime=stop_time)
         input.apply(time)
         fmu.exitInitializationMode()
 
     recorder = Recorder(fmu=fmu, modelDescription=model_description, variableNames=output, interval=output_interval)
 
-    n_steps = 0
+    n_steps = time / output_interval
 
     # simulation loop
-    while time < stop_time:
+    while time + output_interval <= stop_time:
 
         if timeout is not None and (current_time() - sim_start) > timeout:
             break
